/*
Firmware for an auto-baudrate-detector and -convertor implemented in an
ATTiny2313.
(C) 2009 Jeroen Domburg (jeroen AT spritesmods.com)

This program is free software: you can redistribute it and/or modify
t under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.
	    
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
			    
You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "io.h"
#include "uart.h"
#include "timer.h"

#include "avr/interrupt.h"
#include "util/delay.h"
#include <stdio.h>

//Sort tickBuff, from smallest to largest.
static void sortList(void) {
	//Plain ole bubble sort, on so little values that shouldn't take too long :)
	char swapped=1;
	long t;
	while (swapped) {
		swapped=0;
		for (char x=0; x<TICKBUFFLEN-1; x++) {
			if (tickBuff[(int)x]>tickBuff[(int)x+1]) {
				t=tickBuff[(int)x+1];
				tickBuff[(int)x+1]=tickBuff[(int)x];
				tickBuff[(int)x]=t;
				swapped=1;
			}
		}
	}
	//Ok, done :)
}

//Calculate 1/baudrate from the contents of tickBuff.
static long calculateDelay(void) {
	unsigned long proposedDelay;
	unsigned long finalDelayAcc=0;
	//Ok, tickBuff is now filled. Sort it first.
	sortList();
	//Sanity check:
	//The largest time measured can be at most 9 bit-times
	//(1 startbit + 8 zero-valued databits). Refuse if this
	//isn't true by a long shot. (take 15x for a good margin)
	if ((tickBuff[0]*15)<tickBuff[TICKBUFFLEN-1]) return -1;
	//Take the smallest value as the initial proposed delay.
	finalDelayAcc=tickBuff[0];
	proposedDelay=tickBuff[0];
	//Now, figure out the delay.
	for (char x=1; x<TICKBUFFLEN; x++) {
		unsigned long rest=tickBuff[(int)x];
		char n=0;
		while (rest>proposedDelay) {
			n++;
			rest-=proposedDelay;
		}
		//Round up -> if tickbuff=proposedDelay*4.6, n=5.
		if (rest>(proposedDelay/2)) n++;
		
		//Ok, we now know how many bit-times the delay was. We can
		//derive a better bit-time estimation from that. Add that estimation
		//to the accumulative estimations.
		finalDelayAcc+=(tickBuff[(int)x]/n);
	
		//Calculate the new proposed delay
		proposedDelay=finalDelayAcc/(x+1);
	}
	//Ok, proposedDelay now is the best estimate of bit-time we have.
	return proposedDelay;
}

int main(void) {
	initIo();
	timerInit();
	sei();
	while (1) { //eternal loop
		
		//Do autobaud.
		tickPos=0; //Reset tickBuff pointer so tickBuff will be filled by ICP int
		//Sleep until we've collected TICKBUFFLEN samples.
		while (tickPos!=TICKBUFFLEN) ;
		//Ok, samples are in. Calculate delay.
		long delay=calculateDelay();
		if (delay!=-1) { //Only if calculation succeeded.
			long baudRate=F_CPU/(unsigned long)delay;
			//Tell user the baudrate.
			serialPutChar('<');
			serialPutInt(baudRate);
			serialPutChar('B');
			serialPutChar('>');
			serialPutChar('\r');
			serialPutChar('\n');
			uartBase=delay;
			
			//Ok, we're synced. Wait till it's probable we're out of sync
			//(much rx errors) and then resync.
			//Until then, just re-send everything we receive.
			do {
				//Reset error & char counter
				swUartRxErrs=0;	swUartRxChars=0;
				//Wait for 8 chars
				while (swUartRxChars<8) {
					if (uartShouldRecv) {
						//We should receive :)
						int b=swUartRecv();
						uartShouldRecv=0;
						swUartRxChars++;
						if (b==-1) {
							//Error in reception :/
							swUartRxErrs++;
						} else {
							//Reception OK :)
							serialPutChar(b);
						}
					}
				}
				//If more than 2 errors in the 8 bytes, resync.
			} while (swUartRxErrs<3);
		}
	}
}
