/*
Firmware for an auto-baudrate-detector and -convertor implemented in an
ATTiny2313.
(C) 2009 Jeroen Domburg (jeroen AT spritesmods.com)

This program is free software: you can redistribute it and/or modify
t under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.
	    
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
			    
You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "timer.h"
#include "uart.h"
#include <avr/interrupt.h>
#include <avr/io.h>
#include <io.h>

//This is where the magic happens: the timer routines.

volatile unsigned long tickBuff[TICKBUFFLEN];
volatile char tickPos;
volatile char uartShouldRecv;
volatile unsigned int uartBase;

//This is incremented every time timer1 overflows,
//so timerHi:TCNT1 taken together is a 32-bit timer counting
//at the clock frequenct.
volatile unsigned int timerHi;

//True if sw uart is busy receiving
volatile char uartRxing;
//Bits to send over the sw uart
volatile unsigned int uartTxBits;

//Counters: amount of errors in reception and amount of bytes received.
volatile unsigned char swUartRxErrs;
volatile unsigned char swUartRxChars;

//True if the main routine should call swUartRecv() to receive
//a byte from the device.
volatile char uartShouldRecv;
//This contains the time at the middle of the start bit of the
//received byte.
volatile int uartRecvStartTime;


//Tx buffer
volatile unsigned char txBuff[16];
volatile unsigned char txWrPtr;
volatile unsigned char txRdPtr;


void timerInit() {
	tickPos=0;
	TCCR1A=0;
	TCCR1B=0x81; //Input capture noise filter on, trigger on 0, div=1
	TIMSK=0x88; //Tim1 ovf & tim1 input capture int on
	uartBase=0;
	txWrPtr=0;
	txRdPtr=0;
}

//Sets the handshake line according to the status of the tx buffer.
//If the buffer is half empty, cts will be set so the PC won't send any
//more bytes.
#define setHandshaking() pinCtsSet(((txWrPtr-txRdPtr)&15)>8)

//Send the next byte in the sw uart tx buffer. Disable the
//tx routines if the buffer is empty.
static inline void sendNextByte(void) {
	if (txWrPtr!=txRdPtr) {
		txRdPtr++;
		txRdPtr&=15;
		setHandshaking();
		//uartTxBits contains the bits to send:
		//start(0), 8x data, stop(1)
		uartTxBits=(txBuff[txRdPtr]<<1)|0x200;
		TIFR|=0x20; //disable outstanding ints
		TIMSK|=0x20; //enable t1ocb int
		OCR1B=TCNT1+10; //trigger after this bit
	} else {
		TIMSK&=~0x20; //disable t1ocb int
	}
}

//Warning: works only with ints disabled.
//Make the sw uart send a byte.
void swUartSend(unsigned char byte) {
	txWrPtr++; 
	txWrPtr&=15;
	txBuff[txWrPtr]=byte;
	setHandshaking();
	//If tx int is disabled, wake that routine again.
	if ((TIMSK&0x20)==0) sendNextByte();
}

//Software UART TX: This is called as soon as a bit has to be sent.
ISR(TIMER1_COMPB_vect) {
	PORTB|=0x80;
	OCR1B+=uartBase; //Line up for next bit.
	if (uartTxBits==0) { //All done?
		sendNextByte(); //Yes -> fetch next byte, if any.
	} else {
		pinTxSet(uartTxBits&1); //Set TX according to bit
		uartTxBits>>=1; //Nope -> next bit.
	}
}


//We don't want to inline this at any cost: it'll quickly fill up the
//flash of the device.
char uartRecvWaitForBit(void) __attribute__((noinline));
//Wait till the middle of the next bit to be received.
//This routine returns false if the receive pin has been seen to 
//go high somewhere here. If that happens during the start bit,
//the baudrate probably is incorrect.
char uartRecvWaitForBit(void) {
	unsigned int w;
	char ret=1;
	PORTB|=0x80;
	do {
		w=TCNT1-uartRecvStartTime;
		if (pinRxVal()) ret=0;
		if (w&0x8000) {
			//We have time. Allow the ISRs to do their thing.
			sei(); asm("nop"); cli();
		}
	} while (w&0x8000);
	uartRecvStartTime+=uartBase;
	PORTB&=~0x80;
	return ret;
}
//Receive a byte using the sw uart.
int swUartRecv() {
	unsigned char b=0;
	char startok;
	cli(); //Interrupts will be only enabled when we know for sure we can spare the time for them.
	uartRxing=1;
	swUartRxChars++;
	//Get startbit; return -1 if it seemed to be corrupted.
	startok=uartRecvWaitForBit();
	if (!startok) {
		uartRxing=0;
		sei();
		return -1;
	}
	//All ok. Fetch data bits.
	uartRecvWaitForBit();
	if (pinRxVal()) b|=1;
	uartRecvWaitForBit();
	if (pinRxVal()) b|=2;
	uartRecvWaitForBit();
	if (pinRxVal()) b|=4;
	uartRecvWaitForBit();
	if (pinRxVal()) b|=8;
	uartRecvWaitForBit();
	if (pinRxVal()) b|=16;
	uartRecvWaitForBit();
	if (pinRxVal()) b|=32;
	uartRecvWaitForBit();
	if (pinRxVal()) b|=64;
	uartRecvWaitForBit();
	if (pinRxVal()) b|=128;
	uartRxing=0; //stopbit 'doesn't count', if the line goes down here it's ok to start a new byte.
	uartRecvWaitForBit();
	if (!pinRxVal()) {
		sei();
		return -1;
	}
	sei();
	return b;
}

ISR(TIMER1_CAPT_vect) {
	static unsigned long lowStartPos;
	
	if ((TCCR1B&0x40)==0) {
		//Check if we need the sw uart to receive something
		if (!uartRxing && uartBase!=0) {
			uartShouldRecv=1;
			uartRecvStartTime=ICR1+(uartBase>>1);
			//Compensate for icr slowness and code duration.
			//Number is somewhat arbitrary.
			uartRecvStartTime-=50;
		}
		//Don't do anything if no recording is needed.
		if (tickPos==TICKBUFFLEN) return;
		//We triggered on a hi->lo event.
		lowStartPos=ICR1+((unsigned long)timerHi<<16);
		//If we trigger when timer=0, we've missed the timer
		//overflow int because this int has higer priority.
		//Fake its effect here instead.
		if (ICR1==0) lowStartPos+=(1L<<16);
		TCCR1B|=0x40; //Trigger on lo->hi from now on.
	} else {
		//We triggered on a low->hi event.
		
		lowStartPos=(ICR1+((unsigned long)timerHi<<16))-lowStartPos;
		if (ICR1==0) lowStartPos-=(1L<<16);
		//lowStartPos now contains the length of the low pulse.
		//Store and move on.
		if (tickPos!=TICKBUFFLEN) { //look for space first
			tickBuff[(int)(tickPos++)]=lowStartPos; //store
		}
		TCCR1B&=(~0x40); //Trigger on hi->lo from now on.
	}
	//Ok, all done :) If needed, the main routine will handle the autobaud
	//calculations if tickBuff is full.
}

ISR(TIMER1_OVF_vect) {
	timerHi++; //yawn
}

